package hexagon;

import org.joml.Vector3f;

import hexagon.util.MathUtil;
import hexagon.world.World;

public class HexBox {
	private final float radius;
	private final float height;
	private final Vector3f offset;// right angle coords (same as Camera.position)
	
	public HexBox(float radius, float height, Vector3f offset) {
		this.radius = radius;
		this.height = height;
		this.offset = offset;
	}
	
	/** Assuming it doesn't already collide, will it collide after timeStep seconds? */
	public boolean willCollide(Vector3f _position, Vector3f velocity, float timeStep, World world) {
		Vector3f placeInBlock = _position.add(offset, new Vector3f());// temporary storage
		placeInBlock.x += velocity.x * timeStep;
		placeInBlock.y += velocity.y * timeStep;
		placeInBlock.z += velocity.z * timeStep;
		
		Vector3f blockCoords = MathUtil.toBlockCoords(placeInBlock, null, placeInBlock, false);
		Vector3f position = blockCoords.add(placeInBlock, new Vector3f());
		
		int xx = (int) blockCoords.x;
		int zz = (int) blockCoords.z;
		for (int y = (int) Math.floor(position.y + 0.5f), ymax = (int) Math.floor(position.y + 0.5f + height); y <= ymax; y++) {
			for (int x = -1; x <= 1; x++) {
				for (int z = -1; z <= 1; z++) {
					if (x * z != 1) {// corner pieces
						if (intersect(position, xx + x, y, zz + z) && world.getBlock(xx + x, y, zz + z) != null) {
							return true;
						}
					}
				}
			}
		}
		
		return false;
	}
	
	private boolean intersect(Vector3f position, int x, int y, int z) {
		boolean intY = y + 0.5f > position.y && y - 0.5f < position.y + height;
		boolean intX = x + 0.5f > position.x - radius && x - 0.5f < position.x + radius;
		boolean intZ = z + 0.5f > position.z - radius && z - 0.5f < position.z + radius;
		boolean intW = z - x + 0.5f > position.z - position.x - radius && z - x - 0.5f < position.z - position.x + radius;
		
		return intY && intX && intZ;// && intW;
	}
}
