package hexagon.util;

import org.joml.Vector3f;
import org.joml.Vector4f;

import hexagon.Camera;
import hexagon.Main;
import hexagon.model.HexagonModel;
import hexagon.world.Block;
import hexagon.world.World;
import hexagon.world.block.BlockType;

public class MousePicker {
	
	private Vector4f tempVec4;
	private Vector3f currentRay;
	private long hoverBlock;
	private boolean hoverBlockExists = false;
	private int hoverSide;
	
	private Camera camera;
	
	// Used in updateBlockPointedAt
	private Vector3f[] points = new Vector3f[12];
	
	public MousePicker(Camera camera) {
		this.camera = camera;
		tempVec4 = new Vector4f();
		currentRay = new Vector3f();
		
		for (int i = 0; i < 12; i++)
			points[i] = new Vector3f();
	}
	
	public Vector3f getCurrentRay() {
		return currentRay;
	}
	
	public long getHoverBlock() {
		return hoverBlock;
	}
	
	public boolean hoverBlockExists() {
		return hoverBlockExists;
	}
	
	public int getHoverSide() {
		return hoverSide;
	}
	
	public void update() {
		currentRay = calculateMouseRay();
	}
	
	private Vector3f calculateMouseRay() {
		if (Main.main.moveWithMouse) tempVec4.set(0, 0, -1f, 1f);
		else tempVec4.set((float) Main.main.mousePos.x * 2f / Main.main.windowSize.x - 1, (float) Main.main.mousePos.y * 2f / Main.main.windowSize.y - 1, -1f, 1f);
		return toWorldCoords(toEyeCoords(tempVec4));
	}
	
	private Vector3f toWorldCoords(Vector4f eyeCoords) {
		eyeCoords.mul(camera.invViewMatr);
		currentRay.set(eyeCoords.x, eyeCoords.y, eyeCoords.z);
		currentRay.normalize();
		return currentRay;
	}
	
	private Vector4f toEyeCoords(Vector4f clipCoords) {// Reusing clipCoords a lot
		clipCoords.mul(camera.invProjMatr);
		clipCoords.set(clipCoords.x, clipCoords.y, -1f, 0f);
		return clipCoords;
	}
	
	public void updateBlockPointedAt(World world) {
		Vector3f camEucl = Main.main.camera.realPosition.getEucl();
		Vector3f current = new Vector3f(Main.main.camera.blockCoords);
		BlockType block = world.getBlock((int) current.x, (int) current.y, (int) current.z);
		if (block != null) {
			hoverBlock = World.encodeBlockCoords((int) current.x, (int) current.y, (int) current.z);
			hoverBlockExists = true;
			hoverSide = -1;
			return;
		}
		Vector3f PA = new Vector3f(), PB = new Vector3f(), temp = new Vector3f();
		Vec3f tempVec3f = new Vec3f();
		do {
			for (int i = 0; i < 12; i++) {
				double v = i % 6 / 3d * Math.PI;
				temp.set((float) Math.cos(v), 1 - i / 6, (float) (Math.sin(v) / HexagonModel.y60));
				points[i].set(MathUtil.fromBlockCoords(current, temp, tempVec3f).getEucl());
			}
			int index = 0, side = -1;
			points[index + 6].sub(camEucl, PA);
			points[index].sub(camEucl, PB);
			boolean startRight = toTheRight(currentRay, PA, PB);
			if (startRight) {
				index = 6;
				do {
					index--;
					if (index == 0) break;
					points[index + 6].sub(camEucl, PA);
					points[index].sub(camEucl, PB);
				} while (toTheRight(currentRay, PA, PB));
			} else {
				do {
					index++;
					if (index == 6) {
						System.err.println("Index == 6. This is a bug!\n\tcurrent: " + current + "\n\tcurrentRay: " + currentRay);
						return;
					}
					points[index + 6].sub(camEucl, PA);
					points[index].sub(camEucl, PB);
				} while (!toTheRight(currentRay, PA, PB));
				index--;
			}
			points[(index + 1) % 6].sub(camEucl, PA);
			points[index].sub(camEucl, PB);
			if (toTheRight(currentRay, PA, PB)) { // under ceiling
				points[index + 6].sub(camEucl, PA);
				points[(index + 1) % 6 + 6].sub(camEucl, PB);
				if (toTheRight(currentRay, PA, PB)) side = index + 2; // over floor
				else side = 1;
			} else side = 0;
			
			if (side == 0) {
				points[(index + 5) % 6].sub(points[index], PA);
				points[(index + 1) % 6].sub(points[index], PB);
			} else if (side == 1) {
				points[(index + 1) % 6 + 6].sub(points[index + 6], PA);
				points[(index + 5) % 6 + 6].sub(points[index + 6], PB);
			} else {
				points[index].sub(points[index + 6], PA);
				points[(index + 1) % 6 + 6].sub(points[index + 6], PB);
			}
			Vector3f pointOnSide = points[side == 0 ? index : index + 6];
			Vector3f normal = PA.cross(PB, new Vector3f());
			
			if (currentRay.dot(normal) > 0) break;//TODO: this is a temporary fix for rayloops
			
			float distance = Math.abs(pointOnSide.sub(camEucl, new Vector3f()).dot(normal) / currentRay.dot(normal));//abs may be needed (a/-0)
			if (distance <= World.blockToCoordRatio * World.reachDistance) {
				byte[] offsets = Block.neighbourOffsets[side];
				current.set(current.x + offsets[0], current.y + offsets[1], current.z + offsets[2]);
				block = world.getBlock((int) current.x, (int) current.y, (int) current.z);
				if (block != null) {
					hoverBlock = World.encodeBlockCoords((int) current.x, (int) current.y, (int) current.z);;
					hoverSide = side < 2 ? 1 - side : (side + 1) % 6 + 2; // (side - 2 + 3) % 6 + 2
					hoverBlockExists = true;
					return;
				}
				continue;
			}
			break;
		} while (true);
		hoverBlock = 0;
		hoverBlockExists = false;
		return;
	}
	
	private boolean toTheRight(Vector3f ray, Vector3f PA, Vector3f PB) {
		Vector3f cross = PB.cross(ray, new Vector3f());
		float dot = PA.dot(cross);
		return dot > 0;
	}
	
}
