package hexagon.world;

import java.util.Arrays;

public class Cluster {
	protected World world;
	protected Chunk[] chunks;
	private int chunkCount = 0;
	public final long coords;
	
	public Cluster(long coords, World world) {
		this.coords = coords;
		this.world = world;
	}
	
	public Cluster(int x, int y, int z, World world) {
		this(World.encodeClusterCoords(x, y, z), world);
	}
	
	public int x() {
		return (int) (coords >> 24 & 0xffff) << 16 >> 16;
	}
	
	public int y() {
		return (int) (coords >> 16 & 0xff) << 24 >> 24;
	}
	
	public int z() {
		return (int) (coords & 0xffff);
	}
	
	public int size() {
		return chunkCount;
	}
	
	public Chunk getChunk(int chunk) {
		return chunks != null ? chunks[chunk & 0xfff] : null;
	}
	
	public Chunk getChunk(int x, int y, int z) {
		return chunks != null ? chunks[encodeCoords(x, y, z)] : null;
	}
	
	public Chunk setChunk(Chunk chunk) {
		int coords = chunk.coords;
		if (chunks == null) chunks = new Chunk[16 * 16 * 16];
		Chunk c = chunks[coords];
		chunks[coords] = chunk;
		if (c != null) c.onRemove(false);
		else chunkCount++;
		return chunk;
	}
	
	public Chunk removeChunk(int x, int y, int z) {
		int coords = encodeCoords(x, y, z);
		Chunk c = chunks[coords];
		chunks[coords] = null;
		if (c != null) {
			c.onRemove(false);
			chunkCount--;
		}
		return c;
	}
	
	public void onRemove(boolean noUpdate) {
		if (chunks != null) {
			Arrays.fill(chunks, null);
			for (Chunk c : chunks)
				if (c != null) c.onRemove(noUpdate);
		}
		world = null;
	}
	
	public static short encodeCoords(int x, int y, int z) { // bits in short: 4x 4y 4z
		return (short) ((x & 0xf) << 8 | (y & 0xf) << 4 | z & 0xf);
	}
	
	protected void saveChunks() {
		for (Chunk c : chunks)
			if (c != null) c.saveChunk();
	}
}
